@@ -7,14 +7,25 @@ from typing import Optional, List
 class Scanner:
     """Scan the file system for duplicate files"""
 
-    def __init__(self, *, db_path: str, clean=False, prompt=False):
+    def __init__(
+            self, *,
+            db_path: str,
+            clean=False,
+            prompt=False,
+            copy=False,
+            ignore_existing=False,
+    ):
         """
         Scanner for duplicate file detection
         :param db_path: Path to the sqlite3 database file
         :param clean: If true then delete any existing database file before starting
         :param prompt: If true then prompt for confirmation
+        :param copy: If true then copy files instead of moving them
+        :param ignore_existing: If true then ignore existing archive files
         """
         self.prompt = prompt
+        self.copy = copy
+        self.ignore_existing = ignore_existing
         print('xclean: File de-duplication utility')
         print()
         if clean is True:
@@ -69,7 +80,8 @@ class Scanner:
         """
         Scan directory and subdirectories for main files
         :param dir_path: Path to directory to start the scan in
-        :param include: Optional filename extensions to scan
+        :param include: Optional filename extensions to include in scan
+        :param exclude: Optional filename extensions to exclude from scan
         """
         dir_path = os.path.realpath(dir_path)
         print(f'Scan {dir_path} for main files')
@@ -80,15 +92,9 @@ class Scanner:
             if len(eligible_file_names) > 0:
                 dir_id = self._record_main_directory(root_dir)
                 for file_name in eligible_file_names:
-                    file_path = os.path.join(root_dir, file_name)
-                    stat_info = os.stat(file_path)
-                    file_size = stat_info.st_size
                     file_count += 1
+                    file_size = self._record_new_file(dir_id, root_dir, file_name)
                     total_size += file_size
-                    self._cur.execute(
-                        '''INSERT INTO FileInfo (file_size, dir_id, file_name) VALUES (?,?,?)''',
-                        (file_size, dir_id, file_name)
-                    )
             self._con.commit()
         print()
         print(f'{file_count:,} files scanned with {total_size:,} bytes')
@@ -100,6 +106,16 @@ class Scanner:
             }
         }
 
+    def _record_new_file(self, dir_id, root_dir, file_name):
+        file_path = os.path.join(root_dir, file_name)
+        stat_info = os.stat(file_path)
+        file_size = stat_info.st_size
+        self._cur.execute(
+            '''INSERT INTO FileInfo (file_size, dir_id, file_name) VALUES (?,?,?)''',
+            (file_size, dir_id, file_name)
+        )
+        return file_size
+
     def _record_main_directory(self, root_dir: str) -> int:
         dir_id = self._create_directory_entry(root_dir)
         self._remove_old_directory_files(dir_id)
@@ -161,6 +177,7 @@ class Scanner:
             check_aae=False,
             archive_to: Optional[str]=None,
             archive_new: Optional[str]=None,
+            import_new: Optional[str]=None,
             unprotect=False,
             new=False,
             dup=False,
@@ -177,6 +194,7 @@ class Scanner:
         :param check_aae: If true then check the aae matches as well
         :param archive_to: Path to archive duplicate files to
         :param archive_new: Path to archive new files to
+        :param import_new: Path to import new files into
         :param unprotect: If true then unprotect main files
         :param new: If true then report new files
         :param dup: If true then report duplicate files
@@ -265,6 +283,7 @@ class Scanner:
                             target_file_path=target_file_path,
                             dir_path=dir_path,
                             archive_new=archive_new,
+                            import_new=import_new,
                             new_count=new_count,
                             file_size=file_size,
                             new=new,
@@ -302,25 +321,63 @@ class Scanner:
 
     def _handle_new_file(
             self, *,
-            target_file_path, dir_path, archive_new,
+            target_file_path, dir_path, archive_new, import_new,
             new_count, file_size,
             new=False,
     ) -> bool:
+
         if new:
             print()
             print(f'  New  {target_file_path}')
             print(f'       {new_count:,} : (size {file_size:,})')
+
         if archive_new is not None:
-            if not self._archive_file(target_file_path, dir_path, archive_new, archive_type='new'):
+            if not self._archive_file(
+                    target_file_path, dir_path, archive_new,
+                    archive_type='new',
+            ):
                 return False
             xmp_file_path = Scanner._find_xmp_file(target_file_path)
             if xmp_file_path is not None:
-                if not self._archive_file(xmp_file_path, dir_path, archive_new, archive_type='new'):
+                if not self._archive_file(
+                        xmp_file_path, dir_path, archive_new,
+                        archive_type='new',
+                ):
                     return False
             aae_file_path = Scanner._find_aae_file(target_file_path)
             if aae_file_path is not None:
-                if not self._archive_file(aae_file_path, dir_path, archive_new, archive_type='new'):
+                if not self._archive_file(
+                        aae_file_path, dir_path, archive_new,
+                        archive_type='new',
+                ):
                     return False
+
+        if import_new is not None:
+
+            for target_file_path in [
+                target_file_path,
+                Scanner._find_xmp_file(target_file_path),
+                Scanner._find_aae_file(target_file_path),
+            ]:
+                if target_file_path is None:
+                    continue
+                if not self._archive_file(
+                        target_file_path, dir_path, import_new,
+                        archive_type='new',
+                ):
+                    return False
+                archive_file_path = self._archive_file_path(target_file_path, dir_path, import_new)
+                import_new_parts = import_new.split('/')
+                archive_file_parts = archive_file_path.split('/')
+                dir_id = None
+                root_dir = None
+                for parts in range(len(import_new_parts), len(archive_file_parts)):
+                    root_dir = '/'.join(archive_file_parts[0:parts])
+                    dir_id = self._create_directory_entry(root_dir)
+                file_name = archive_file_parts[-1]
+                self._record_new_file(dir_id, root_dir, file_name)
+            self._con.commit()
+
         return True
 
     def _handle_duplicate_file(
@@ -337,15 +394,21 @@ class Scanner:
             print(f'  Dup  {target_file_path}')
             print(f'       {dups_count:,} : (size {file_size:,})')
         if archive_to is not None:
-            if not self._archive_file(target_file_path, dir_path, archive_to):
+            if not self._archive_file(
+                    target_file_path, dir_path, archive_to,
+            ):
                 return False
             xmp_file_path = Scanner._find_xmp_file(target_file_path)
             if xmp_file_path is not None:
-                if not self._archive_file(xmp_file_path, dir_path, archive_to):
+                if not self._archive_file(
+                        xmp_file_path, dir_path, archive_to,
+                ):
                     return False
             aae_file_path = Scanner._find_aae_file(target_file_path)
             if aae_file_path is not None:
-                if not self._archive_file(aae_file_path, dir_path, archive_to):
+                if not self._archive_file(
+                        aae_file_path, dir_path, archive_to,
+                ):
                     return False
         elif trash_dups is True:
             if not self._trash_file(target_file_path):
@@ -464,10 +527,7 @@ class Scanner:
             archive_to: str,
             archive_type='duplicate',
     ) -> bool:
-        target_file_suffix = target_file_path[len(dir_path):]
-        while target_file_suffix.startswith('/'):
-            target_file_suffix = target_file_suffix[1:]
-        archive_file_path = os.path.join(archive_to, target_file_suffix)
+        archive_file_path = self._archive_file_path(target_file_path, dir_path, archive_to)
         archive_dir_path = os.path.dirname(archive_file_path)
         print(f'  Archive {archive_type} file to {archive_file_path}')
         if not os.path.exists(target_file_path):
@@ -475,6 +535,9 @@ class Scanner:
             return False
         if os.path.exists(archive_file_path):
             print(f'Archive file already exists {archive_file_path}')
+            if self.ignore_existing:
+                if os.stat(archive_file_path).st_size == os.stat(target_file_path).st_size:
+                    return True
             return False
         if not self._prompted():
             return False
@@ -490,12 +553,21 @@ class Scanner:
         if os.stat(archive_file_path).st_size != os.stat(target_file_path).st_size:
             print(f'Failed to fully copy {target_file_path} to {archive_file_path}')
             return False
-        os.remove(target_file_path)
-        if os.path.exists(target_file_path):
-            print(f'Failed to remove {target_file_path}')
-            return False
+        if not self.copy:
+            print(f'  Remove {target_file_path}')
+            os.remove(target_file_path)
+            if os.path.exists(target_file_path):
+                print(f'Failed to remove {target_file_path}')
+                return False
         return True
 
+    def _archive_file_path(self, target_file_path: str, dir_path: str, archive_to: str) -> str:
+        target_file_suffix = target_file_path[len(dir_path):]
+        while target_file_suffix.startswith('/'):
+            target_file_suffix = target_file_suffix[1:]
+        archive_file_path = os.path.join(archive_to, target_file_suffix)
+        return archive_file_path
+
     @staticmethod
     def _files_are_the_same(source_file_path: str, target_file_path: str, check_xmp=False, check_aae=False) -> bool:
         source_fp = os.open(source_file_path, os.O_RDONLY)
@@ -543,34 +615,27 @@ class Scanner:
 
     @staticmethod
     def _find_xmp_file(file_path: str) -> Optional[str]:
-        xmp_file_path = f'{file_path}.xmp'
-        if os.path.exists(xmp_file_path):
-            return xmp_file_path
-        xmp_file_path = f'{file_path}.XMP'
-        if os.path.exists(xmp_file_path):
-            return xmp_file_path
-        prefix, extn = os.path.splitext(file_path)
-        xmp_file_path = f'{prefix}.xmp'
-        if os.path.exists(xmp_file_path):
-            return xmp_file_path
-        xmp_file_path = f'{prefix}.XMP'
-        if os.path.exists(xmp_file_path):
-            return xmp_file_path
-        return None
+        return Scanner._find_file(file_path, 'xmp')
 
     @staticmethod
     def _find_aae_file(file_path: str) -> Optional[str]:
-        aae_file_path = f'{file_path}.aae'
+        return Scanner._find_file(file_path, 'aae')
+
+    @staticmethod
+    def _find_file(file_path: str, extn_lower: str):
+        extn_lower = extn_lower.lower()
+        extn_upper = extn_lower.upper()
+        aae_file_path = f'{file_path}.{extn_lower}'
         if os.path.exists(aae_file_path):
             return aae_file_path
-        aae_file_path = f'{file_path}.AAE'
+        aae_file_path = f'{file_path}.{extn_upper}'
         if os.path.exists(aae_file_path):
             return aae_file_path
         prefix, extn = os.path.splitext(file_path)
-        aae_file_path = f'{prefix}.aae'
+        aae_file_path = f'{prefix}.{extn_lower}'
         if os.path.exists(aae_file_path):
             return aae_file_path
-        aae_file_path = f'{prefix}.AAE'
+        aae_file_path = f'{prefix}.{extn_upper}'
         if os.path.exists(aae_file_path):
             return aae_file_path
         return None
